@@ -1,30 +1,33 @@
+use v5.10;
+
 package Mac::PropertyList;
 use strict;
 
 use warnings;
 no warnings;
 
-use vars qw($ERROR $XML_head $XML_foot $VERSION @EXPORT_OK %EXPORT_TAGS);
+use vars qw($ERROR);
 use Carp qw(croak carp);
 use Data::Dumper;
 use XML::Entities;
 
-use parent qw(Exporter);
+use Exporter qw(import);
 
-@EXPORT_OK = qw(
+our @EXPORT_OK = qw(
 	parse_plist
 	parse_plist_fh
 	parse_plist_file
 	plist_as_string
 	create_from_hash
 	create_from_array
+	create_from_string
 	);
 
-%EXPORT_TAGS = (
+our %EXPORT_TAGS = (
 	'all' => \@EXPORT_OK,
 	);
 
-$VERSION = '1.41_01';
+our $VERSION = '1.603_02';
 
 =encoding utf8
 
@@ -59,9 +62,10 @@ Mac::PropertyList - work with Mac plists at a low level
 =head1 DESCRIPTION
 
 This module is a low-level interface to the Mac OS X Property List
-(plist) format. You probably shouldn't use this in applications–build
-interfaces on top of this so you don't have to put all the heinous
-multi-level object stuff where people have to look at it.
+(plist) format in either XML or binary. You probably shouldn't use
+this in applications–build interfaces on top of this so you don't have
+to put all the heinous multi-level object stuff where people have to
+look at it.
 
 You can parse a plist file and get back a data structure. You can take
 that data structure and get back the plist as XML. If you want to
@@ -70,6 +74,14 @@ change the structure inbetween that's your business. :)
 You don't need to be on Mac OS X to use this. It simply parses and
 manipulates a text format that Mac OS X uses.
 
+If you need to work with the old ASCII or newer JSON formet, you can
+use the B<plutil> tool that comes with MacOS X:
+
+	% plutil -convert xml1 -o ExampleBinary.xml.plist ExampleBinary.plist
+
+Or, you can extend this module to handle those formats (and send a pull
+request).
+
 =head2 The Property List format
 
 The MacOS X Property List format is simple XML. You can read the DTD
@@ -103,6 +115,7 @@ There are several types of objects:
 	Mac::PropertyList::data
 	Mac::PropertyList::real
 	Mac::PropertyList::integer
+	Mac::PropertyList::uid
 	Mac::PropertyList::date
 	Mac::PropertyList::array
 	Mac::PropertyList::dict
@@ -144,23 +157,12 @@ information or blessed objects.
 
 my $Debug = $ENV{PLIST_DEBUG} || 0;
 
-$XML_head =<<"XML";
-<?xml version="1.0" encoding="UTF-8"?>
-<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
-<plist version="1.0">
-XML
-
-$XML_foot =<<"XML";
-</plist>
-XML
-
 my %Readers = (
 	"dict"    => \&read_dict,
 	"string"  => \&read_string,
 	"date"    => \&read_date,
 	"real"    => \&read_real,
 	"integer" => \&read_integer,
-	"string"  => \&read_string,
 	"array"   => \&read_array,
 	"data"    => \&read_data,
 	"true"    => \&read_true,
@@ -178,6 +180,8 @@ will be imported unless you ask for it.
 
 	use Mac::PropertyList qw( :all );
 
+=head2 Things that parse
+
 =over 4
 
 =item parse_plist( TEXT )
@@ -194,6 +198,7 @@ sub parse_plist {
 
 	my $plist = do {
 		if( $text =~ /\A<\?xml/ ) { # XML plists
+			$text =~ s/<!--(?:[\d\D]*?)-->//g;
 			# we can handle either 0.9 or 1.0
 			$text =~ s|^<\?xml.*?>\s*<!DOC.*>\s*<plist.*?>\s*||;
 			$text =~ s|\s*</plist>\s*$||;
@@ -260,7 +265,7 @@ sub parse_plist_file {
 		return;
 		}
 
-	my $text = do { local $/; open my($fh), $file; <$fh> };
+	my $text = do { local $/; open my($fh), '<:raw', $file; <$fh> };
 
 	parse_plist( $text );
 	}
@@ -279,9 +284,12 @@ Returns a string representing the hash in the plist format.
 sub create_from_hash {
 	my $hash  = shift;
 
-	return unless UNIVERSAL::isa( $hash, 'HASH' );
+	unless( ref $hash eq ref {} ) {
+		carp "create_from_hash did not get an hash reference";
+		return;
+		}
 
-	my $string = "$XML_head" . Mac::PropertyList::dict->write_open . "\n";
+	my $string = XML_head() . Mac::PropertyList::dict->write_open . "\n";
 
 	foreach my $key ( keys %$hash ) {
 		next if ref $hash->{$key};
@@ -296,7 +304,7 @@ sub create_from_hash {
 		$string .= $bit;
 		}
 
-	$string .= Mac::PropertyList::dict->write_close . "\n$XML_foot";
+	$string .= Mac::PropertyList::dict->write_close . "\n" . XML_foot();
 
 	return $string;
 	}
@@ -315,9 +323,12 @@ Returns a string representing the array in the plist format.
 sub create_from_array {
 	my $array  = shift;
 
-	return unless UNIVERSAL::isa( $array, 'ARRAY' );
+	unless( ref $array eq ref [] ) {
+		carp "create_from_array did not get an array reference";
+		return;
+		}
 
-	my $string = "$XML_head" . Mac::PropertyList::array->write_open . "\n";
+	my $string = XML_head() . Mac::PropertyList::array->write_open . "\n";
 
 	foreach my $element ( @$array ) {
 		my $value = Mac::PropertyList::string->new( $element );
@@ -328,11 +339,52 @@ sub create_from_array {
 		$string .= $bit;
 		}
 
-	$string .= Mac::PropertyList::array->write_close . "\n$XML_foot";
+	$string .= Mac::PropertyList::array->write_close . "\n" . XML_foot();
 
 	return $string;
 	}
 
+=item create_from_string( STRING )
+
+Returns a string representing the string in the plist format.
+
+=cut
+
+sub create_from_string {
+	my $string  = shift;
+
+	unless( ! ref $string ) {
+		carp "create_from_string did not get a string";
+		return;
+		}
+
+	return
+		XML_head() .
+		Mac::PropertyList::string->new( $string )->write .
+		"\n" . XML_foot();
+	}
+
+=item create_from
+
+Dispatches to either C<create_from_array>, C<create_from_hash>, or
+C<create_from_string> based on the argument. If none of those fit,
+this C<croak>s.
+
+=cut
+
+sub create_from {
+	my $thingy  = shift;
+
+	return do {
+		if(      ref $thingy eq ref [] ) { &create_from_array  }
+		elsif(   ref $thingy eq ref {} ) { &create_from_hash   }
+		elsif( ! ref $thingy eq ref {} ) { &create_from_string }
+		else {
+			croak "Did not recognize argument! Must be a string, or reference to a hash or array";
+			}
+		};
+	}
+
 =item read_string
 
 =item read_data
@@ -373,11 +425,12 @@ sub read_next {
 	while( not defined $value ) {
 		croak "Couldn't read anything!" if $source->eof;
 		$_ .= $source->get_line;
-
-		if( s[^\s* < (string|date|real|integer|data) >
-			   \s*(.*?)\s* </\1> ][]sx ) {
+		if( s[^\s* < (string|date|real|integer|data) > \s*(.*?)\s* </\1> ][]sx ) {
 			$value = $Readers{$1}->( $2 );
 			}
+		elsif( s[^\s* < string / > ][]x ){
+			$value = $Readers{'string'}->( '' );
+			}
 	    elsif( s[^\s* < (dict|array) > ][]x ) {
 			# We need to put back the unprocessed text if
 			# any because the <dict> and <array> readers
@@ -432,7 +485,13 @@ sub read_dict {
 		}
 
 	$source->put_line( $_ );
-	return Mac::PropertyList::dict->new( \%hash );
+	if ( 1 == keys %hash && exists $hash{'CF$UID'} ) {
+	    # This is how plutil represents a UID in XML.
+	    return Mac::PropertyList::uid->integer( $hash{'CF$UID'}->value );
+	    }
+	else {
+	    return Mac::PropertyList::dict->new( \%hash );
+	    }
 	}
 
 =item read_array
@@ -467,6 +526,38 @@ sub read_data {
 	return Mac::PropertyList::data->new( $string );
 	}
 
+=back
+
+=head2 Things that write
+
+=over 4
+
+=item XML_head
+
+Returns a string that represents the start of the PList XML.
+
+=cut
+
+sub XML_head () {
+	<<"XML";
+<?xml version="1.0" encoding="UTF-8"?>
+<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
+<plist version="1.0">
+XML
+	}
+
+=item XML_foot
+
+Returns a string that represents the end of the PList XML.
+
+=cut
+
+sub XML_foot () {
+	<<"XML";
+</plist>
+XML
+	}
+
 =item plist_as_string
 
 Return the plist data structure as XML in the Mac Property List format.
@@ -476,11 +567,11 @@ Return the plist data structure as XML in the Mac Property List format.
 sub plist_as_string {
 	my $object = CORE::shift;
 
-	my $string = $XML_head;
+	my $string = XML_head();
 
 	$string .= $object->write . "\n";
 
-	$string .= $XML_foot;
+	$string .= XML_foot();
 
 	return $string;
 	}
@@ -495,6 +586,10 @@ is really just C<as_perl>.
 
 sub plist_as_perl { $_[0]->as_perl }
 
+=back
+
+=cut
+
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 package Mac::PropertyList::Source;
 sub new {
@@ -507,13 +602,34 @@ sub eof { (not @{$_[0]->{buffer}}) and $_[0]->source_eof }
 sub get_line {
 	my $self = CORE::shift;
 
+# I'm not particularly happy with what I wrote here, but that's why
+# you shouldn't write your own buffering code! I might have left over
+# text in the buffer. This could be stuff a higher level looked at and
+# put back with put_line. If there's text there, grab that.
+#
+# But here's the tricky part. If that next part of the text looks like
+# a "blank" line, grab the next next thing and append that.
+#
+# And, if there's nothing in the buffer, ask for more text from
+# get_source_line. Follow the same rules. IF you get back something that
+# looks like a blank line, ask for another and append it.
+#
+# This means that a returned line might have come partially from the
+# buffer and partially from a fresh read.
+#
+# At some point you should have something that doesn't look like a
+# blank line and the while() will break out. Return what you do.
+#
+# Previously, I wasn't appending to $_ so newlines were disappearing
+# as each next read replaced the value in $_. Yuck.
+
 	local $_ = '';
 	while (defined $_ && /^[\r\n\s]*$/) {
 		if( @{$self->{buffer}} ) {
-			$_ = shift @{$self->{buffer}};
+			$_ .= shift @{$self->{buffer}};
 			}
 		else {
-			$_ = $self->get_source_line;
+			$_ .= $self->get_source_line;
 			}
 		}
 
@@ -559,8 +675,6 @@ sub value {
 sub type { my $r = ref $_[0] ? ref $_[0] : $_[0]; $r =~ s/.*:://; $r; }
 
 sub new {
-	#print STDERR "Got [@_]\n";
-
 	bless $_[1], $_[0]
 	}
 
@@ -654,8 +768,6 @@ package Mac::PropertyList::dict;
 use base qw(Mac::PropertyList::Container);
 
 sub new {
-	#print STDERR Data::Dumper::Dumper( $_[1] );
-
 	$_[0]->SUPER::new( $_[1] );
 	}
 
@@ -757,6 +869,74 @@ use base qw(Mac::PropertyList::Scalar);
 package Mac::PropertyList::integer;
 use base qw(Mac::PropertyList::Scalar);
 
+# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
+package Mac::PropertyList::uid;
+use base qw(Mac::PropertyList::Scalar);
+
+# The following is conservative, since the actual largest unsigned
+# integer is ~0, which is 0xFFFFFFFFFFFFFFFF on many (most?) modern
+# Perls; but it is consistent with Mac::PropertyList::ReadBinary.
+# This is really just future-proofing though, since it appears from
+# CFBinaryPList.c that a UID is carried as a hard-coded uint32_t.
+use constant LONGEST_HEX_REPRESENTABLE_AS_NATIVE => 8;	# 4 bytes
+
+# Instantiate with hex string. The string will be padded on the left
+# with zero if its length is odd. It is this string which will be
+# returned by value(). Presence of a non-hex character causes an
+# exception. We default the argument to '00'.
+sub new {
+	my ( $class, $value ) = @_;
+	$value = '00' unless defined $value;
+	Carp::croak( 'uid->new() argument must be hexadecimal' )
+		if $value =~ m/ [[:^xdigit:]] /smx;
+	substr $value, 0, 0, '0'
+		if length( $value ) % 2;
+	return $class->SUPER::new( $value );
+	}
+
+# Without argument, this is an accessor returning the value as an unsigned
+# integer, either a native Perl value or a Math::BigInt as needed.
+# With argument, this is a mutator setting the value to the hex
+# representation of the argument, which must be an unsigned integer,
+# either native Perl of Math::BigInt object. If called as static method
+# instantiates a new object.
+sub integer {
+	my ( $self, $integer ) = @_;
+	if ( @_ < 2 ) {
+		my $value = $self->value();
+		return length( $value ) > LONGEST_HEX_REPRESENTABLE_AS_NATIVE ?
+			Math::BigInt->from_hex( $value ) :
+			hex $value;
+		}
+	else {
+		Carp::croak( 'uid->integer() argument must be unsigned' )
+			if $integer < 0;
+		my $value = ref $integer ?
+			$integer->to_hex() :
+			sprintf '%x', $integer;
+		if ( ref $self ) {
+			substr $value, 0, 0, '0'
+				if length( $value ) % 2;
+			${ $self } = $value;
+			}
+		else {
+			$self = $self->new( $value );
+			}
+		return $self;
+		}
+	}
+
+# This is how plutil represents a UID in XML.
+sub write {
+	my $self = shift;
+	my $dict = Mac::PropertyList::dict->new( {
+			'CF$UID' => Mac::PropertyList::integer->new(
+				$self->integer ),
+			}
+		);
+	return $dict->write();
+	}
+
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 package Mac::PropertyList::string;
 use base qw(Mac::PropertyList::Scalar);
@@ -809,13 +989,11 @@ package Mac::PropertyList::false;
 use base qw(Mac::PropertyList::Boolean);
 
 
-=back
-
 =head1 SOURCE AVAILABILITY
 
 This project is in Github:
 
-	git://github.com/briandfoy/mac-propertylist.git
+	https://github.com/briandfoy/mac-propertylist.git
 
 =head1 CREDITS
 
@@ -841,14 +1019,16 @@ all the plist junk and just play with the data.
 
 =head1 AUTHOR
 
-brian d foy, C<< <bdfoy@cpan.org> >>
+brian d foy, C<< <briandfoy@pobox.com> >>
+
+Tom Wyant added support for UID types.
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright © 2004-2015, brian d foy <bdfoy@cpan.org>. All rights reserved.
+Copyright © 2004-2025, brian d foy <briandfoy@pobox.com>. All rights reserved.
 
 This program is free software; you can redistribute it and/or modify
-it under the same terms as Perl itself.
+it under the terms of the Artistic License 2.0.
 
 =head1 SEE ALSO
 
